﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectで手による操作を行うNUI: Natural User Interfaceの実験』
//【名称】Kinect_ProximitySensor　　【Version】SDK Ver.1.6対応版
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章  Copyright by ☆Tomoaki Ueda☆
//【作成日】2012年11月05日
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Media;                   //【SoundPlayer】
using System.Diagnostics;             //【Debug】
using System.Drawing;                 //【Bitmap】【Color】
using System.Drawing.Imaging;         //【BitmapData】
using System.Runtime.InteropServices; //【Marshal.Copy】
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択

namespace Kinect_ProximitySensor
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private ColorImageFrame KS_ColorFrame;  //【KINECTのColorImageFrame】
        private byte[] KS_ColorImageDataXRGB;   //【KINECTのColorImageData】32bit/画素,X8R8G8B8表示形式
        private byte[] KS_InfraredImageData;    //【KINECTのColorImageData】16bit/画素,GrayScale表示形式
        private int KS_ColorFrameWidth;         //【KINECTのColorFrameの横方向画素数】
        private int KS_ColorFrameHeight;        //【KINECTのColorFrameの縦方向画素数】
        private Bitmap KS_bmp1;                 //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;            //【ColorImage用BitmapData】
        private Rectangle KS_rect1;             //【ColorImage表示枠】
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        private int BitShift = 0;
        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private NTL_ColorBar NTL_InfraredBar= new NTL_ColorBar();
        private Point NTL_Finger; //【指先の座標】
        private int NTL_Infrared; //【赤外線強度】
        private const int NTL_ITh = 62500; //【Ok判断の近接位置の赤外線強度】
        private int[] NTL_Rect = { 40, 90, 370, 300 }; //【判定領域】x0, y0, w, h
        private int NTL_Phase;    //【コマンド・フェーズ】0:待機(I<Th) 1:Ok 
        private SoundPlayer[] NTL_Sound = new SoundPlayer[3]; //【音】
        private string[] NTL_WaveFile = { "Ok.wav", "NG.wav", "Done.wav" }; //【音ファイル】
        private string[] NTL_Command = { "Green.jpg", "Red.jpg", "Ok.jpg", "Right.jpg", "Left.jpg", "Up.jpg", "Down.jpg" }; //【画像】
        private int tCount = 0; //【表示を待機色Greenに戻すためのカウンタ】
        private const int NTL_nSpan = 20;
        private const double NTL_nSpan1 = 1.0 / (double)NTL_nSpan;
        private const double NTL_nSpan2 = 1.0 / (double)(NTL_nSpan * NTL_nSpan);
        private const int NTL_nSpanOfs = NTL_nSpan >> 1;
        private int[] NTL_FIFOx;  //【1次元水平方向FIFO、サイズはNTL_nSpan】
        private int[] NTL_FIFOxy; //【2次元矩形FIFO、サイズはNTL_nSpan*KS_ColorFrameWidth】
        private int[] NTL_SumXY;  //【矩形領域移動平均、サイズはKS_ColorFrameWidth】
        private int NTL_Count; //【表示用カウンタ】
        private Boolean NTL_flag;

        public Form1()
        {
            InitializeComponent();

            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors 
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();
            //【検出結果の表示】
            if (KS_KinectSensor == null) //【Kinect接続が確認できなかった場合】
            {
                pictureBox2.Image = imageList1.Images[1];
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            else //【Kinect検出に成功】
            {
                pictureBox2.Image = imageList1.Images[0];
            }

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】ColorStreamを許可
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.InfraredResolution640x480Fps30); //【InfraredStream許可】

            //【KINECT動作開始】
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】
            KS_KinectSensor.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(ColorImageReady); //【ColorImageReadyをcallするよう設定】
            KS_ColorImageDataXRGB = new byte[2*KS_KinectSensor.ColorStream.FramePixelDataLength]; //【カラー画像データ１フレーム分を保存するメモリを確保】
            KS_InfraredImageData = new byte[KS_KinectSensor.ColorStream.FramePixelDataLength];    //【赤外線画像データ１フレーム分を保存するメモリを確保】
            KS_ColorFrameWidth = KS_KinectSensor.ColorStream.FrameWidth;   //【ColorImageの横方向画素数】
            KS_ColorFrameHeight = KS_KinectSensor.ColorStream.FrameHeight; //【ColorImageの縦方向画素数】
            KS_bmp1 = new Bitmap(KS_ColorFrameWidth, KS_ColorFrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【ColorImage表示用】
            KS_rect1 = new Rectangle(0, 0, KS_ColorFrameWidth, KS_ColorFrameHeight); //【ColorImage用枠】
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【キャプチャー画像がpictureBox枠内にはめ込まれます】

            //【Kinectヘッドの仰角を設定】設定範囲は±27°
            int Angle = 0; //【初期化直後のKinectの仰角は0度[水平]】
            KS_KinectSensor.ElevationAngle = Angle;
            textBox1.Text = Angle.ToString();

            NTL_InfraredBar.CreateColorPalletForInfrared(65536); //【カラーパレットを16bit用で準備】
            pictureBox3.Image = NTL_InfraredBar.CreateColorScale(400, 10);

            //【FIFO初期化】
            NTL_FIFOx = new int[NTL_nSpan];
            NTL_FIFOxy = new int[NTL_nSpan * KS_ColorFrameWidth];
            NTL_SumXY = new int[KS_ColorFrameWidth];

            NTL_Count = 0; //【表示用カウンタの初期化】
            NTL_flag = false; //【移動平均表示をオフにする】

            NTL_Phase = 0; //【待機】
            for (int i = 0; i < NTL_Sound.Length; i++) { NTL_Sound[i] = new SoundPlayer("../../Wave/" + NTL_WaveFile[i]); } //【音初期化】

            this.FormClosing += new FormClosingEventHandler(Form1_FormClosing);
            checkBox1.CheckedChanged += new EventHandler(checkBox1_CheckedChanged);
            checkBox2.CheckedChanged += new EventHandler(checkBox2_CheckedChanged);
            checkBox3.CheckedChanged += new EventHandler(checkBox3_CheckedChanged);
        }

        //***************************************************************
        //【終了処理】
        //***************************************************************
        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
        }

        //***************************************************************
        //【ColorImageの取得イベント】
        //***************************************************************
        // Kinectからカラー画像１フレーム分が到着する度に、この関数が呼ばれます。
        // Windowsパソコンが最も高速にBitmapデータを表示できるデータ形式は1画素４バイトの形式なのでXRGB形式を使っています。
        // 但し、追加のCG描画で半透明表現を使うときはARGB形式を使います。
        private void ColorImageReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            int ix, iy, img, add, maximg;
            int i, L;
            int SumX, pX, pXY, pY, nF;

            using (KS_ColorFrame = e.OpenColorImageFrame())
            {
                if (KS_ColorFrame != null)
                {
                    KS_ColorFrame.CopyPixelDataTo(KS_InfraredImageData); //【KINECTのColor画像を配列KS_InfraredImageDataにコピー】

                    //【移動平均用FIFOの初期化】
                    for (i = 0; i < NTL_nSpan; i++) { NTL_FIFOx[i] = 0; } //【FIFOxの0クリア】
                    for (i = 0; i < KS_ColorFrameWidth * NTL_nSpan; i++) { NTL_FIFOxy[i] = 0; }
                    for (i = 0; i < KS_ColorFrameWidth; i++) { NTL_SumXY[i] = 0; }
                    SumX = 0; pX = pXY = pY = 0;
                    nF = NTL_nSpan * KS_ColorFrameWidth;

                    ix = iy = maximg = 0; //【最大輝度を示す画素のスクリーン座標を求める処理の初期化】
                    L = 0;                //【グレースケールをカラーに変更する処理の管理ポインタの初期化】
                    for (i = 0; i < KS_InfraredImageData.Length; i += 2)
                    {
                        img = KS_InfraredImageData[i+1] * 256 + KS_InfraredImageData[i]; //【IRイメージ取得】

                        //【FIFOxの移動平均処理】【水平方向】【スキャンライン】
                        SumX += img - NTL_FIFOx[pX];       //【差分計算から累積加算値を求める】
                        NTL_FIFOx[pX] = img;               //【新規データをFIFOに保存する】
                        pX++; if (pX == NTL_nSpan) pX = 0; //【FIFOx用ポインタ(0～n-1)の更新処理】

                        //【FIFOxyの移動平均処理】【垂直方向】【スクリーン横幅分並列処理】
                        NTL_SumXY[pY] += SumX - NTL_FIFOxy[pXY];    //【差分計算から累積加算値を求める】
                        NTL_FIFOxy[pXY] = SumX;                     //【新規データをFIFOに保存する】
                        add = NTL_SumXY[pY];                        //【n×n画素の矩形領域の累積加算値⇒add】
                        pY++; if (pY == KS_ColorFrameWidth) pY = 0; //【SumXY用ポインタ(0～width-1)の更新処理】
                        pXY++; if (pXY == nF) pXY = 0;              //【FIFOxy用ポインタ(0～n*width-1)の更新処理】

                        //【値が最大値ならxy座標を保存】
                        if ((ix>NTL_nSpan)&&(maximg < add))
                        {
                            maximg = add;
                            NTL_Finger.X = ix;
                            NTL_Finger.Y = iy;
                        }
                        //【スクリーン座標を更新】
                        ix++;
                        if (ix == KS_ColorFrameWidth)
                        {
                            ix = 0;
                            iy++;
                        }

                        //【画素値変数img(16bitグレースケール)をカラールックアップテーブル(LUT)を使って着色表示する】
                        if (NTL_flag) { img = (int)((double)add * NTL_nSpan2); } //【矩形領域平均値を表示する場合有効にする】
                        img = (img << BitShift) & 0xffff; //【BitShiftは通常0で使用する】
                        KS_ColorImageDataXRGB[L] = NTL_InfraredBar.NTL_ColorElement[4 * img];
                        KS_ColorImageDataXRGB[L + 1] = NTL_InfraredBar.NTL_ColorElement[4 * img + 1];
                        KS_ColorImageDataXRGB[L + 2] = NTL_InfraredBar.NTL_ColorElement[4 * img + 2];
                        KS_ColorImageDataXRGB[L + 3] = NTL_InfraredBar.NTL_ColorElement[4 * img + 3];
                        L += 4;
                    }
                    KS_data1 = KS_bmp1.LockBits(KS_rect1, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
                    Marshal.Copy(KS_ColorImageDataXRGB, 0, KS_data1.Scan0, KS_ColorImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp1.UnlockBits(KS_data1);                                           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp1;                                            //【pictureBox1に画像表示】

                    Graphics g = Graphics.FromImage(pictureBox1.Image); //【pixtureBox1にグラフィックスを上書きする為の準備】
                    ix = NTL_Finger.X; //【最大輝度を示すスクリーン座標x】
                    iy = NTL_Finger.Y; //【最大輝度を示すスクリーン座標y】
                    if (KS_flagFlip) { ix = KS_ColorFrameWidth - ix; }  //【左右反転表示のための処理】
                    Color lc = Color.Azure;                             //【領域位置表示色の定義】
                    g.DrawLine(new Pen(lc, 3.0f), ix + NTL_nSpanOfs, 0, ix + NTL_nSpanOfs, 480); //【画面中央縦線】
                    g.DrawLine(new Pen(lc, 3.0f), 0, iy - NTL_nSpanOfs, 640, iy - NTL_nSpanOfs); //【画面中央横線】
                    g.Dispose(); //【描画終了】
                    Validate();  //【更新を反映する】
                    NTL_Infrared = (int)(maximg * NTL_nSpan2); //【検出した最大輝度値を表示用変数Infraredに渡す】
                    DisplayLabels();       //【数値表示】6フレームに1回表示

                    if (NTL_Infrared > NTL_ITh) //【近接位置】
                    {
                        switch(NTL_Phase)
                        {
                            case 0:
                                if ((NTL_Finger.X > NTL_Rect[0]) && (NTL_Finger.X < (NTL_Rect[0] + NTL_Rect[2]))
                                && (NTL_Finger.Y > NTL_Rect[1]) && (NTL_Finger.Y < (NTL_Rect[1] + NTL_Rect[3])))
                                {
                                   NTL_Action(1, 2, 0,"受付中"); //【Okポジション内】phase:1, 2:Ok.jpg, 0:Ok.wav
                                }
                                break;

                            case 1:
                                if ((NTL_Finger.X > NTL_Rect[0]) && (NTL_Finger.X < (NTL_Rect[0] + NTL_Rect[2])))
                                {
                                   if(NTL_Finger.Y <= NTL_Rect[1])
                                   {
                                      NTL_Action(2, 5, 2, "上方"); //【上方】phase:2(実行中), 5:Up.jpg, 2:Done.wav
                                      //【ここに上方指示時の処理を記述します】
                                   }
                                   if (NTL_Finger.Y >= (NTL_Rect[1] + NTL_Rect[3]))
                                   {
                                      NTL_Action(2, 6, 2, "下方"); //【下方】phase:2(実行中), 6:Down.jpg, 2:Done.wav
                                      //【ここに下方指示時の処理を記述します】
                                   }
                                }
                                if ((NTL_Finger.Y > NTL_Rect[1]) && (NTL_Finger.Y < (NTL_Rect[1] + NTL_Rect[3])))
                                {
                                   if (NTL_Finger.X >= (NTL_Rect[0] + NTL_Rect[2]))
                                   {
                                      NTL_Action(2, 3, 2, "右方"); //【右方】phase:2(実行中), 3:Right.jpg, 2:Done.wav
                                      //【ここに右方指示時の処理を記述します】
                                   }
                                   if (NTL_Finger.X <= NTL_Rect[0])
                                   {
                                      NTL_Action(2, 4, 2, "左方"); //【左方】phase:2(実行中), 4:Left.jpg, 2:Done.wav
                                      //【ここに左方指示時の処理を記述します】
                                   }
                                }
                                break;

                            default:
                                break;
                        }
                    }
                    else //【近接と判断されない】
                    {
                        if (NTL_Phase == 1) { NTL_Action(2, 2, 1, "無効"); }
                    }
                    tCount++;
                    if ((tCount == 60) && (NTL_Phase == 2)) { NTL_Action(0, 0, -1, "待機中"); } //【２秒後に待機状態に復帰させます】
                }
                else
                {
                    //【ImageFrameがnullだったのは要求が間に合わなかったと考えられます】
                }
            }
        }
       
        //***************************************************************
        //【フェーズ・イメージ・音・テキストの表示処理】p:フェーズ　im:イメージ　sd:音　t:状態をテキストで表示
        //***************************************************************
        private void NTL_Action(int p, int im, int sd, string t)
        {
            NTL_Phase = p;                 //【管理フェーズを設定】
            pictureBox4.Image = Bitmap.FromFile("../../image/" + NTL_Command[im]); //【イメージを表示】
            if (sd >= 0) //【sdが-1なら無音】
            {
                NTL_Sound[sd].LoadAsync(); //【非同期に音をロード】
                NTL_Sound[sd].Play();      //【音を再生する】
            }
            if (p == 2) { tCount = 0; }    //【tCountをクリア】★60(2秒経過)で待機状態に復帰させるためのカウンタ
            label5.Text = t;               //【状態をテキストで表示】
        }

        //***************************************************************
        //【指定フレーム数毎に最近点座標とKionix３軸直交加速度センサの測定値を表示】
        //***************************************************************
        private void DisplayLabels()
        {
            NTL_Count++;        //【表示用カウンタを+1】
            if (NTL_Count == 6) //【30FPSなら0.2秒に1回表示】
            {
                NTL_Count = 0;  //【表示用カウンタをクリア】

                label2.Text = "x座標=" + NTL_Finger.X.ToString();
                label3.Text = "y座標=" + NTL_Finger.Y.ToString();
                label4.Text = "強度I=" + NTL_Infrared.ToString();

                Vector4 Acc = KS_KinectSensor.AccelerometerGetCurrentReading();
                label9.Text = "x=" + Acc.X.ToString() + "[g]";
                label10.Text = "y=" + Acc.Y.ToString() + "[g]";
                label11.Text = "z=" + Acc.Z.ToString() + "[g]";
                label12.Text = "|Acc|=" + Math.Sqrt(Acc.X * Acc.X + Acc.Y * Acc.Y + Acc.Z * Acc.Z).ToString() + "[g]";
            }
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 angle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if (angle <= KS_KinectSensor.MaxElevationAngle && angle >= KS_KinectSensor.MinElevationAngle)
            {
                KS_KinectSensor.ElevationAngle = angle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        //　ここではフラグを立てるだけで、実際の処理はColorImageReady内で行います。
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【プロジェクタの照射のOn/Offを切り替える】
        //***************************************************************
        private void checkBox2_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox2.Checked) { KS_KinectSensor.ForceInfraredEmitterOff = true; }
            else { KS_KinectSensor.ForceInfraredEmitterOff = false; }
        }

        //***************************************************************
        //【移動平均表示のOn/Offを切り替える】
        //***************************************************************
        void checkBox3_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox3.Checked) { NTL_flag = true; } else { NTL_flag = false; }
        }

        //***************************************************************
        //【指定されたビットシフト量を読込む⇒増感量】
        //***************************************************************
        private void button1_Click(object sender, EventArgs e)
        {
            int i = int.Parse(textBox2.Text);
            if (i < 0 || i > 8) return;
            BitShift = i;
        }
    }
}
